#
# -------------------------------------------------------------------------
# Get link magin data depending on modcod from Ayecka SR1 & format for MRTG
# -------------------------------------------------------------------------
#
#
# Copyright (C) 2014-2015 Francis Breame
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# For details of the GNU General Public License see
# http://perldoc.perl.org/perlgpl.html, or write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
#
# Author:	Francis Breame G8ISI
# 		vf0123@btinternet.com
#
#
# 0.1.0	140709	- Initial issue.
# 0.1.1 140710	- Remove reported IP.
#		  Blank negative values.
# 0.1.2	140906	- Report unknown values as UNKNOWN to avoid MTRG errors.
# 0.1.3 140910	- When starting, file is not written until both values have been read 
#		  or 2 mins have elapsed at fast sampling rate. Any existing file will
#		  therefore be maintained.
# 0.1.4 150118	- Improved error debugging.
# 0.1.5 151026	- Better error handling.
#		  Report zero values if error.
#		  Normal mode force uses time rather than cycle counts.
#
# 170714 - version just for 16APSK2/3

use strict;
no strict "subs";
use warnings;
no warnings "reserved";

use Net::SNMP;
use Time::HiRes qw (usleep);

#--------------------------------------------
my $title 	= "SR1lm v0.1.5 151026\n";
#--------------------------------------------

my ($ip, $rx, $debug, $fn, $fo, $session, $error, $sample, $singleSNMP, $vals);
my ($startTime, $fileTime, $starting, $j);
my (@varbindlist, $modcod, $lastModcod, $lm, $lm16, $lastLm16, $uptime);

my $Q = '"';

# SNMP values
my $community		= "public";
my $version		= 2;

my $lmOID		= "1.3.6.1.4.1.27928.101.1.+.4.6.0";
my $modcodStatusOID	= "1.3.6.1.4.1.27928.101.1.+.4.7.0";
my $uptimeOID		= "1.3.6.1.2.1.1.3.0";

my $dvbs2_8psk_3_5	= 12;
my $dvbs2_16apsk_2_3	= 18;

my $startupSample	= 0.5;	# Secs
my $normalSample	= 5;

my $startupForceFinish	= 120;	# No of secs in startup mode after which normal mode is forced
my $normalForceWrite	= 300;	# No of secs in normal mode after which file write is forced

my $errorValue		= 0;	# This will be reported to MRTG as lm16 + uptime to MRTG 
				# in the event of an error
my $modcodErrorValue	= -99;	# Mustn't clash with legit values (-1 is legit)

my $unknownValue	= "UNKNOWN";


sub getTS
{
	my ($s,$mi,$h,$d,$m,$y) = localtime(time);
	return sprintf "%04d/%02d/%02d %02d:%02d:%02d", $y+1900,$m+1,$d,$h,$mi,$s;
}

sub debug
{
	return if (!$debug);

	my $s = shift;

	print STDERR getTS()." ".$s;
}

sub openSession
{
	debug("Opening SNMP session to $ip community=$community version=$version\n");

	undef $session;		# Destroy any old object in the event of previous errors

	# Blocking mode by default
	($session, $error) = Net::SNMP->session(
                           -hostname      => $ip,
                           -version       => $version,
                           -community     => $community 
                        );

	if (defined $session)
	{
		debug("Opened OK\n");
	}
	else
	{
		debug("Failed - error=$error\n");
	}
}

sub init
{
	$modcod = $lm = $uptime = $lm16 = $unknownValue;
	$lastLm16 = $lastModcod = $unknownValue;
	$startTime = $fileTime = time;
	$sample = $startupSample;		# Fast rate to start with
	$starting = 1;
}

sub getSingleRequest
{
	# Sets $error & returns specified error if error

	my ($oid, $errRet) = @_;

	my ($val);

	$val = $session->get_request($oid);
	if (defined $val)
	{
		return $$val{$oid};
	}
	else
	{
		$error = $session->error();
		return $errRet;
	}
}

sub normalise
{
	# Normalise value reported to MRTG
	# 	- $errorValue in place of -ve or blank value

	my $n = shift;

	return ($n eq $unknownValue || ($n ne "" && $n>=0)) ? $n : $errorValue;
}

sub writeFile
{
	open $fo, ">$fn" || die "Can't open output file '$fn'" if (defined $fn); 
				# Reset to overwrite as long it's an actual file 
	print $fo normalise($lm16)."\n".normalise($lm16)."\n$uptime\nAyecka SR1\n";
	close $fo if (defined $fn); 

	debug("File written: age=".(time-$fileTime)."s modcod=$modcod/$lastModcod ".
		"lm16=$lm16/$lastLm16\n");

	$fileTime = time;
}


# Default args
$ip	    = shift;
$rx	    = 1;
$singleSNMP = 0;
$debug	    = 0;

if (!defined $ip || $ip !~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/)
{
	print STDERR "Invalid or missing IP or output file.\n".
		"Usage: (w)perl SR1lm.pl <IP> [-f<output file>] [-r<rx no>] [-s<sample secs>] [-b<n>] [-h<n>] [-o]\n".
		"e.g. (w)perl SR1lm.pl 192.168.0.249 ${Q}-fC:\\MRTG 2\\mib\\SR1lm.txt$Q -s10\n".
		"See installation notes for details.\n\n".
		"Important: perl library Net::SNMP must be installed.\n";
	exit;
}

# Parse args
while (defined ($_=shift))
{
	/^-f/io  && do {s/^-f//io;  $fn = $_;  next};			# Output file
	/^-r/io  && do {s/^-r//io;  $rx = $_;  next};			# RX no
	/^-s/io  && do {s/^-s//io;  $normalSample = $_;  next};		# Sample rate s
	/^-b/io  && do {s/^-b//io;  $dvbs2_8psk_3_5 = $_;  next};	# Basic modcod value
	/^-h/io  && do {s/^-h//io;  $dvbs2_16apsk_2_3 = $_;  next};	# High volume modcod value
	/^-o/io  && do {	    $singleSNMP = 1;  next};		# Single SNMP request mode

	/^-d/io  && do {	    $debug = 1;  next};
	print STDERR "Invalid argument '$_'\n"; 
}

if ($rx<1 || $rx>2)
{
	print STDERR "Invalid rx number - must be 1 or 2. Using 1\n";
	$rx = 1;
}

$fo = *STDOUT;		# Use STDOUT in the absence of a file spec ($fn undef)

debug($title);
debug("Single SNMP request mode\n") if ($singleSNMP);
debug("modcod values: basic=$dvbs2_8psk_3_5, HV=$dvbs2_16apsk_2_3\n");

$modcodStatusOID =~ s/\+/$rx/;
$lmOID		 =~ s/\+/$rx/;
push @varbindlist, ($modcodStatusOID, $lmOID, $uptimeOID);

openSession();
init();

# Init file only if it doesn't exist
writeFile() if (defined $fn && !-f $fn);

while (1)
{
	if (defined $session)
	{
		# Errors result in a zero value being returned

		$error  = "";		# Clear any errors

		if ($singleSNMP)
		{

			$modcod = getSingleRequest($modcodStatusOID, $modcodErrorValue);
			$lm     = getSingleRequest($lmOID, $errorValue);
			$uptime = getSingleRequest($uptimeOID, $errorValue);
		}
		else
		{
			$vals = $session->get_request(-varbindlist => \@varbindlist);
			if (defined $vals)
			{
				$modcod = $$vals{$modcodStatusOID};
				$lm     = $$vals{$lmOID};
				$uptime = $$vals{$uptimeOID};
			}
			else
			{
				$error = $session->error();
				$lm = $uptime = $errorValue;
				$modcod = $modcodErrorValue;
			}
		}

		#debug("Raw data: error=$error modcod=$modcod lm=$lm uptime=$uptime\n");

		# NB: modcod can legitimately return -1. In this case, don't change $lm*
		if ($modcod == $dvbs2_16apsk_2_3)
		{
			$lm16 = $lm;
		}
		elsif ($modcod == $modcodErrorValue)
		{
			$lm16 = $errorValue;
		}

		# NB: for */last* comparisons, use ne because of string values initially

		if ($debug && $modcod>=0 && $modcod ne $lastModcod)
		{
			debug("modcod changed $lastModcod -> $modcod\n");
			$lastModcod = $modcod;
		}

		# In normal mode, write triggered by either lm changing or 5 mins elapsed since last write
		if (!$starting)
		{
			if ($lm16 ne $lastLm16 || (time - $fileTime) >= $normalForceWrite)
			{
				writeFile();
				$lastLm16 = $lm16;
			}
		}
		# In startup mode, both lms must have changed or 2 mins elapsed
		elsif (($lm16 ne $lastLm16) || (time - $startTime) >= $startupForceFinish)
		{
			writeFile();
			$lastLm16 = $lm16;
			$starting = 0;
			$sample = $normalSample;
			debug("Normal mode entered\n");
		}
		
		# Error action
		if ($error ne "")
		{
			debug("get_request error: $error\n");
			openSession();	# Have another go
		}
	}
	else
	{
		openSession();		# Keep trying until we get a connection
	}

	usleep($sample * 1000000);
}

